(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = global || self, global.Thingworx = factory());
}(this, (function () { 'use strict';

  /**
   * Class representing an entity collection.
   */
  class EntityCollection {
    /**
     * Creates an entity collection.
     *
     * @param {string} name - the name of the collection.
     * @param {Function} collectionClass - the class of a single entity of the collection.
     * @param {Server} server - the server to which this collection belongs.
     */
    constructor(name, collectionClass, server) {
      this.name = name;
      this.CollectionClass = collectionClass;
      this.server = server;
      this.entities = new Map();
    }

    /**
     * Retrieves the reference to an entity from the collection.
     * If the reference hasn't been retrieved before, adds it to the collection and then returns it.
     *
     * @param {string} name - the name of the entity.
     * @returns {Entity} the entity object.
     */
    getEntity(name) {
      if (!this.entities.has(name)) {
        this.entities.set(name, new this.CollectionClass(name, this));
      }
      return this.entities.get(name);
    }
  }

  /**
   * This class does not extend the Request interface because the Request interface's properties are
   * read-only, and we still need to modify the options of the request in order to put the
   * authentication parameters. When calling .send(), the connection object will add the
   * authentication parameters, and for that, the options need to be modifiable.
   */
  class RequestNeedingAuthentication {
    /**
     * Create a request that needs to be filled in with authentication parameters.
     *
     * @param {string} url - the complete URL this request will go to.
     * @param {object} options - the options for this request.
     * @param {Server} server - the instance of the server this request belongs to.
     */
    constructor(url, options, server) {
      if (!Object.isExtensible(options)) {
        throw new Error('Options of the request cannot be read-only, they need to be writable so the authentication parameters get added');
      }
      this.url = url;
      this.options = options;
      this.server = server;
    }
  }

  /**
   * Base class that represents a response gotten from a Thingworx server.
   */
  class ThingworxResponse {
    /**
     * Shells a JavaScript response object in a ThingworxResponse object.
     *
     * @param {Response} response - a Response object.
     * @param {Request} request - the request the response belongs to.
     */
    constructor(response, request) {
      this.response = response;
      this.request = request;
      this.ok = response.ok; // for easy access, instead of having to do .response.ok from the outside
      this.serverURL = request.server.origin;
    }

    /**
     * Returns asynchronously the body of the response as a string.
     *
     * @returns {Promise} - a promise that resolves to the body of the response.
     */
    async text() {
      return this.response.text();
    }

    /**
     * Returns asynchronously the body of the response as JSON.
     *
     * @returns {Promise} - a promise that resolves to the body of the response.
     */
    async json() {
      return this.response.json();
    }

    /**
     * Returns a string representation of this response including status code and description.
     *
     * @returns {string} the string representation of this response.
     */
    toString() {
      return `${this.response.status} ${this.response.statusText}`;
    }

    /**
     * Creates an error object when the response's request couldn't be completed successful.
     * Only call this method if the .ok property of the response is false.
     *
     * @returns {Error} the error object that was built.
     */
    buildError() {
      return new Error(`Could not connect to the server - Error ${this.toString()}`);
    }
  }

  /**
   * Checks if a property is in an object
   *
   * @param {object} object - the object.
   * @param {*} property - the property to look for in the object.
   * @returns {boolean} - true if the property is in the object, false otherwise.
   */
  function hasProp(object, property) {
    return Object.prototype.hasOwnProperty.call(object, property);
  }

  /**
   * Checks whether a value is an object.
   *
   * @param {*} value - the value to be checked.
   * @returns {boolean} - whether the value is an object.
   */
  function isObject(value) {
    return typeof value === 'object' && value !== null;
  }

  /**
   * Wrapper for a promise that represents a Thingworx response whose body contains JSON.
   */
  class JsonResponsePromise {
    /**
     * Wraps a fetch promise to be handled as request's response whose body contains JSON.
     *
     * @param {Promise} promise - the promise resulting from the fetch to Thingworx.
     * @param {ServiceExecutionRequest|PropertyGetRequest} request - the request that created the
     * promise.
     */
    constructor(promise, request) {
      this.promise = promise;
      this.request = request;
    }

    /**
     * Extracts the single value of the response body.
     * ThingWorx REST API calls always return json objects or infotable-shaped jsons, even when the
     * result is a single value. This method gets the single value we're interested in from the
     * response.
     *
     * @returns {Promise} a promise resolving to the single value we're interested in
     * @throws {Error} an error if the request was unsuccessful or the response is not a single-value
     * result, e.g., the infotable returned contains multiple columns or multiple rows, or the json
     * does not have the shape of an infotable.
     */
    async val() {
      const infoTable = await this.json();
      if (!JsonResponsePromise.isInfoTable(infoTable) || infoTable.rows.length !== 1) {
        throw new Error(`The request ${this.request.toString()} did not return a single-value result. It returned:\n${JsonResponsePromise.prettifyInfoTable(infoTable)}.`);
      }

      const fieldNames = Object.keys(infoTable.dataShape.fieldDefinitions);
      if (fieldNames.length !== 1) {
        throw new Error(`The request ${this.request.toString()} did not return a single-value result. It returned:\n${JsonResponsePromise.prettifyInfoTable(infoTable)}.`);
      }

      const [fieldName] = fieldNames;
      const result = infoTable.rows[0][fieldName];
      const { baseType } = infoTable.dataShape.fieldDefinitions[fieldName];
      return JsonResponsePromise.parseValue(result, baseType);
    }

    /**
     * Parses the response's json body of the fetch request.
     *
     * @returns {Promise} a promise resolving to the object resulting from the json parsing.
     * @throws {Error} if the request was not successful.
     */
    async json() {
      const response = new ThingworxResponse(await this.promise, this.request);
      if (!response.ok) {
        throw await this.buildError(response);
      }
      return response.json();
    }

    /**
     * Builds an error associated to this response.
     * Only call this method if the response's ok status was false.
     *
     * @param {ThingworxResponse} response - the response this promise resolved to.
     * @returns {Error} - an Error object that is ready to be thrown.
     */
    async buildError(response) {
      return new Error(`Error requesting ${this.request.toString()}\nFrom ${response.serverURL}\n${await response.text()}`);
    }

    /* This method is to make this class a Thenable.
       The presence of this method allows this class to be used in conjunction with the await keyword
       without having to use the .json(), .val(), or .infoTable() methods.
    */
    then(successCallback, failureCallback) {
      // The .defaultValue() is implemented in the subclasses.
      // It defines the default value to which this promise should resolve without needing to call the
      // .json(), .val(), or .infoTable() methods.
      return this.defaultValue().then(successCallback, failureCallback);
    }

    /* Implement catch promise method */
    catch(failureCallback) {
      return this.promise.catch(failureCallback);
    }

    /* Implement finally promise method */
    finally(finallyCallback) {
      return this.promise.finally(finallyCallback);
    }

    /**
     * Checks if a value is an infotable.
     *
     * @param {*} value - value to be checked.
     * @returns {boolean} true if value is infotable, false otherwise.
     */
    static isInfoTable(value) {
      return isObject(value)
            && Array.isArray(value.rows)
            && isObject(value.dataShape)
            && isObject(value.dataShape.fieldDefinitions);
    }

    /**
     * Prettifies the infotable by stringifying it to JSON and putting indentation.
     *
     * @param {object} infoTable - an infotable-shaped object.
     * @returns {string} - the string prettified version of the infotable.
     */
    static prettifyInfoTable(infoTable) {
      return JSON.stringify(infoTable, undefined, 2);
    }

    /**
     * Converts a single value received from a Thingworx response JSON to a JavaScript value.
     *
     * @param {number|string|boolean|null} value - the value to be parsed
     * @param {string} baseType - the Thingworx base type of the value.
     *
     * @returns {*} the final value already converted
     */
    static parseValue(value, baseType) {
      switch (baseType) {
        case 'DATETIME':
          return new Date(value);
        case 'INFOTABLE':
          return JsonResponsePromise.parseInfoTable(value);
        default:
          return value;
      }
    }

    /**
     * Converts all JSON values in the infotable to JavaScript values.
     * InfoTables are send through the network as JSONs, which in the end, they are just simple
     * strings. JSON.parse() converted those strings back to JavaScript values, but some values need
     * more conversion. In particular, if you stringify a Date and then parse it as a json, the result
     * will be a string. You still need to convert the string back to a real instance Date. This
     * method does exactly that.
     * Precondition: the infotable object must be valid. JsonResponsePromise.isInfoTable(infoTable)
     * must return true.
     *
     * @param {object} infoTable - the infotable whose values are to be parsed.
     * @returns {object} - the same object given as parameter, for ease of use. The parameter might
     * have been imperatively modified by this method.
     */
    static parseInfoTable(infoTable) {
      const { fieldDefinitions } = infoTable.dataShape;

      const dateTimeFields = [];
      const infoTableFields = [];
      Object.entries(fieldDefinitions).forEach(([fieldName, fieldObj]) => {
        const { baseType } = fieldObj;
        if (baseType === 'DATETIME') {
          dateTimeFields.push(fieldName);
        } else if (baseType === 'INFOTABLE') {
          infoTableFields.push(fieldName);
        }
      });

      infoTable.rows.forEach((row) => {
        dateTimeFields.forEach((dateTimeField) => {
          if (hasProp(row, dateTimeField) && row[dateTimeField] !== null) {
            row[dateTimeField] = new Date(row[dateTimeField]);
          }
        });
        infoTableFields.forEach((infoTableField) => {
          if (hasProp(row, infoTableField) && row[infoTableField] !== null) {
            row[infoTableField] = JsonResponsePromise.parseInfoTable(row[infoTableField]);
          }
        });
      });

      return infoTable;
    }
  }

  /**
   * Wrapper for a promise that represents the result of a property retrieval request.
   */
  class PropertyGetPromise extends JsonResponsePromise {
    /**
     * Uses the .val() method to get the default value when applying the .then() callback on this
     * instance.
     *
     * @returns {Promise} - a promise resolving to the single value of the response.
     */
    defaultValue() {
      return this.val();
    }
  }

  /**
   * Class representing a property request.
   */
  class PropertyGetRequest extends RequestNeedingAuthentication {
    /**
     * Creates a request to retrieve the value of a property.
     *
     * @param {Property} property - the property whose value is to be got.
     */
    constructor(property) {
      const url = PropertyGetRequest.buildRequestURL(property);
      const options = PropertyGetRequest.buildRequestOptions();
      const { server } = property.entity.collection;
      super(url, options, server);
      this.property = property;
    }

    /**
     * Sends the request to a Connection instance so that it sends it to the Thingworx server.
     * Unlike the send method of other requests objects, this one is not asynchronous, because we
     * want to be able to call operations on the promise even before it resolves. Those operations
     * are the methods of the JsonResponsePromise and PropertyGetPromise classes.
     *
     * @param {Connection} connection - the connection instance to use to make the request.
     * @returns {PropertyGetPromise} an instance of PropertyGetPromise.
     */
    send(connection) {
      const promise = connection.sendRequest(this);
      return new PropertyGetPromise(promise, this);
    }

    /**
     * Returns the url for the property get request.
     *
     * @param {Property} property - the property for which to build the URL.
     * @returns {string} - the url for the property get request.
     */
    static buildRequestURL(property) {
      return `${property.entity.collection.server.origin}/Thingworx/${property.entity.collection.name}/${property.entity.name}/Properties/${property.name}`;
    }

    /**
     * Returns the options for the request.
     *
     * @returns {object} - the options used for the request.
     */
    static buildRequestOptions() {
      return {
        method: 'GET',
        headers: {
          Accept: 'application/json',
          'Content-Type': 'application/json',
        },
      };
    }

    /**
     * Builds a string representation of the property get request.
     *
     * @returns {string} - a string representation of the property get request.
     */
    toString() {
      return this.property.toString();
    }
  }

  /**
   * Class representing a Property Set response.
   */
  class PropertySetResponse extends ThingworxResponse {
    /**
     * Builds an error associated to this response.
     * The reason this method is asynchronous is because the response's body contains the error
     * message, and we need to wait until we download it.
     * Only call this method if the response's ok status was false.
     *
     * @returns {Promise} - an object that resolves to an error ready to be thrown.
     */
    async buildError() {
      return new Error(`Error setting ${this.request.toString()}.\nFrom ${this.serverURL}\n${await this.response.text()}`);
    }
  }

  /**
   * Class representing a property set request.
   */
  class PropertySetRequest extends RequestNeedingAuthentication {
    /**
     * Creates a request to update the value of a property.
     *
     * @param {Property} property - the property whose value is to be set.
     * @param {*} value - the new value to set.
     */
    constructor(property, value) {
      const url = PropertySetRequest.buildRequestURL(property);
      const options = PropertySetRequest.buildRequestOptions(property, value);
      const { server } = property.entity.collection;
      super(url, options, server);
      this.property = property;
      this.value = value;
    }

    /**
     * Sends the request to a Connection instance so that it sends it to the Thingworx server.
     *
     * @param {Connection} connection - the connection instance to use to make the request.
     * @returns {PropertySetResponse} an instance of PropertySetResponse representing the response.
     */
    async send(connection) {
      const response = await connection.sendRequest(this);
      return new PropertySetResponse(response, this);
    }

    /**
     * Returns the url for the property set request.
     *
     * @param {Property} property - the property for which to build the URL.
     * @returns {string} - the url for the property set request.
     */
    static buildRequestURL(property) {
      return `${property.entity.collection.server.origin}/Thingworx/${property.entity.collection.name}/${property.entity.name}/Properties/${property.name}`;
    }

    /**
     * Returns the options for the request.
     *
     * @param {Property} property - the property.
     * @param {*} value - the new value of the property.
     * @returns {object} - the options used for the request.
     */
    static buildRequestOptions(property, value) {
      return {
        method: 'PUT',
        headers: {
          Accept: 'application/json',
          'Content-Type': 'application/json',
        },
        body: JSON.stringify({ [property.name]: value }),
      };
    }

    /**
     * Builds a string representation of the property set request.
     *
     * @returns {string} - a string representation of the property set request.
     */
    toString() {
      return `${this.property.toString()} = ${JSON.stringify(this.value, undefined, 2)}`;
    }
  }

  /**
   * Class representing a reference to a property in a Thingworx entity.
   */
  class Property {
    /**
     * Creates a reference to a property in a Thingworx entity.
     *
     * @param {string} name - the name of the property.
     * @param {Entity} entity - the entity to which the property belongs.
     */
    constructor(name, entity) {
      this.name = name;
      this.entity = entity;
    }

    /**
     * Executes a PropertyGetRequest and returns the response.
     *
     * @param {Connection} connection - the connection object to use for the request.
     * @returns {PropertyGetPromise} - an instance of PropertyGetPromise.
     */
    get(connection) {
      const request = new PropertyGetRequest(this);
      return request.send(connection);
    }

    /**
     * Sets the property value by making an HTTP request to the Thingworx server.
     *
     * @param {*} value - the new value of the property.
     * @param {Connection} connection - the connection object to use for the request.
     * @returns {PropertySetResponse} - an instance of PropertySetResponse.
     */
    async set(value, connection) {
      const request = new PropertySetRequest(this, value);
      return request.send(connection);
    }

    /**
     * Builds a string representation of the property.
     *
     * @returns {string} - a string representation of the property.
     */
    toString() {
      return `${this.entity.toString()}.${this.name}`;
    }
  }

  /**
   * Wrapper for a promise that represents the result of a service execution request.
   */
  class ServiceExecutionPromise extends JsonResponsePromise {
    /**
     * Converts the response's json body to an infotable.
     *
     * @returns {Promise} a promise resolving to the infotable of the response body.
     * @throws {Error} if the request was not successful or if the response is not an infotable.
     */
    async infoTable() {
      const infoTable = await this.json();
      if (!JsonResponsePromise.isInfoTable(infoTable)) {
        throw new Error(`The request ${this.request.toString()} did not return an infotable. It returned ${JsonResponsePromise.prettifyInfoTable(infoTable)}.`);
      }
      return JsonResponsePromise.parseInfoTable(infoTable);
    }

    /**
     * Checks that the response is ok and defines undefined as the default value when applying the
     * .then() callback on this instance. This is used only for services that don't have any return
     * value. For example:
     *
     * await Resources['EntityServices'].CreateThing({ ... });
     *
     * @returns {Promise} a promise that resolves to undefined.
     */
    async defaultValue() {
      const response = new ThingworxResponse(await this.promise, this.request);
      if (!response.ok) {
        throw await this.buildError(response);
      }
      return undefined;
    }
  }

  /**
   * Class representing a service execution request.
   */
  class ServiceExecutionRequest extends RequestNeedingAuthentication {
    /**
     * Creates a request to make a execution service Rest API call.
     *
     * @param {Service} service - the service to be called in the request.
     * @param {object} params - the arguments of the service.
     */
    constructor(service, params) {
      const url = ServiceExecutionRequest.buildRequestURL(service);
      const options = ServiceExecutionRequest.buildRequestOptions(params);
      const { server } = service.entity.collection;
      super(url, options, server);
      this.service = service;
      this.params = params;
    }

    /**
     * Sends the request to a Connection instance so that it sends it to the Thingworx server.
     * Unlike the send method of other requests objects, this one is not asynchronous, because we
     * want to be able to call operations on the promise even before it resolves. Those operations
     * are the methods of the JsonResponsePromise and ServiceExecutionPromise classes.
     *
     * @param {Connection} connection - the connection instance to use to make the request.
     * @returns {ServiceExecutionPromise} an instance of ServiceExecutionPromise that will resolve
     * to the response of the service execution.
     */
    send(connection) {
      const promise = connection.sendRequest(this);
      return new ServiceExecutionPromise(promise, this);
    }

    /**
     * Builds a string representation of the service call.
     *
     * @returns {string} - a string representation of the service call.
     */
    toString() {
      const paramsStringified = this.params == null ? '' : JSON.stringify(this.params, undefined, 2);
      return `${this.service.toString()}(${paramsStringified})`;
    }

    /**
     * Returns the url for the service execution request.
     *
     * @param {Service} service - the service for which to build the URL.
     * @returns {string} - the url for the service execution request.
     */
    static buildRequestURL(service) {
      return `${service.entity.collection.server.origin}/Thingworx/${service.entity.collection.name}/${service.entity.name}/Services/${service.name}`;
    }

    /**
     * Returns the options for the request.
     *
     * @param {object} [params] - the parameters of the service.
     * @returns {object} - the options used for the request.
     */
    static buildRequestOptions(params) {
      return {
        method: 'POST',
        headers: {
          Accept: 'application/json',
          'Content-Type': 'application/json',
        },
        body: JSON.stringify(params),
      };
    }
  }

  /**
   * Class representing a reference to a service in a ThingWorx entity.
   */
  class Service {
    /**
     * Creates a reference to a service in a Thingworx entity.
     *
     * @param {string} name - the name of the service.
     * @param {Entity} entity - the entity to which the service belongs.
     */
    constructor(name, entity) {
      this.name = name;
      this.entity = entity;
    }

    /**
     * Calls the service by making an HTTP request to the Thingworx server.
     *
     * @param {object} params - the parameters to be used for the service call.
     * @param {Connection} connection - the connection instance to use to make the HTTP request.
     * @returns {ServiceExecutionPromise} - an instance of ServiceExecutionPromise representing
     * the response that will be get (or was already got) from the server.
     */
    call(params, connection) {
      const request = new ServiceExecutionRequest(this, params);
      return request.send(connection);
    }

    /**
     * Builds a string representation of the service.
     *
     * @returns {string} - a string representation of the service.
     */
    toString() {
      return `${this.entity.toString()}.${this.name}`;
    }
  }

  /**
   * Class representing an entity in ThingWorx.
   */
  class Entity {
    /**
     * Creates a reference to an entity in a Thingworx server.
     *
     * @param {string} name - the name of the entity.
     * @param {EntityCollection} collection - the collection to which this entity belongs.
     */
    constructor(name, collection) {
      this.name = name;
      this.collection = collection;
      this.properties = new Map();
      this.services = new Map();
    }

    /**
     * Gets the reference to a property in the entity.
     * If it doesn't exist yet, adds it to the entity's list of properties and then returns it.
     *
     * @param {string} propertyName - the name of the property.
     * @returns {Property} the property object.
     */
    getProperty(propertyName) {
      if (!this.properties.has(propertyName)) {
        this.properties.set(propertyName, new Property(propertyName, this));
      }
      return this.properties.get(propertyName);
    }

    /**
     * Gets the reference to a service in the entity.
     * If it doesn't exist yet, adds it to the entity's list of services and then returns it.
     *
     * @param {string} serviceName - the name of the service.
     * @returns {Service} the service object.
     */
    getService(serviceName) {
      if (!this.services.has(serviceName)) {
        this.services.set(serviceName, new Service(serviceName, this));
      }
      return this.services.get(serviceName);
    }

    /**
     * Builds a string representation of the entity.
     *
     * @returns {string} - a string representation of the entiy.
     */
    toString() {
      return `${this.collection.name}['${this.name}']`;
    }
  }

  class Dashboard extends Entity {}

  /**
   * Class representing the Dashboards collection in Thingworx.
   */
  class Dashboards extends EntityCollection {
    /**
     * Creates the Dashboards collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('Dashboards', Dashboard, server);
    }
  }

  class DataShape extends Entity {}

  /**
   * Class representing the DataShapes collection in Thingworx.
   */
  class DataShapes extends EntityCollection {
    /**
     * Creates the DataShapes collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('DataShapes', DataShape, server);
    }
  }

  class DataTag extends Entity {}

  /**
   * Class representing the DataTags collection in Thingworx.
   */
  class DataTags extends EntityCollection {
    /**
     * Creates the DataTags collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('DataTags', DataTag, server);
    }
  }

  class Group extends Entity {}

  /**
   * Class representing the Groups collection in Thingworx.
   */
  class Groups extends EntityCollection {
    /**
     * Creates the Groups collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('Groups', Group, server);
    }
  }

  class Log extends Entity {}

  /**
   * Class representing the Logs collection in Thingworx.
   */
  class Logs extends EntityCollection {
    /**
     * Creates the Logs collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('Logs', Log, server);
    }
  }

  class Mashup extends Entity {}

  /**
   * Class representing the Mashups collection in Thingworx.
   */
  class Mashups extends EntityCollection {
    /**
     * Creates the Mashups collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('Mashups', Mashup, server);
    }
  }

  class Menu extends Entity {}

  /**
   * Class representing the Menus collection in Thingworx.
   */
  class Menus extends EntityCollection {
    /**
     * Creates the Menus collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('Menus', Menu, server);
    }
  }

  class ModelTag extends Entity {}

  /**
   * Class representing the ModelTags collection in Thingworx.
   */
  class ModelTags extends EntityCollection {
    /**
     * Creates the ModelTags collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('ModelTags', ModelTag, server);
    }
  }

  class Network extends Entity {}

  /**
   * Class representing the Networks collection in Thingworx.
   */
  class Networks extends EntityCollection {
    /**
     * Creates the Networks collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('Networks', Network, server);
    }
  }

  class Resource extends Entity {}

  /**
   * Class representing the Resources collection in Thingworx.
   */
  class Resources extends EntityCollection {
    /**
     * Creates the Resources collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('Resources', Resource, server);
    }
  }

  class Thing extends Entity {}

  /**
   * Class representing the Things collection in Thingworx.
   */
  class Things extends EntityCollection {
    /**
     * Creates the Things collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('Things', Thing, server);
    }
  }

  class ThingShape extends Entity {}

  /**
   * Class representing the ThingShapes collection in Thingworx.
   */
  class ThingShapes extends EntityCollection {
    /**
     * Creates the ThingShapes collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('ThingShapes', ThingShape, server);
    }
  }

  class ThingTemplate extends Entity {}

  /**
   * Class representing the ThingTemplates collection in Thingworx.
   */
  class ThingTemplates extends EntityCollection {
    /**
     * Creates the ThingTemplates collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('ThingTemplates', ThingTemplate, server);
    }
  }

  class User extends Entity {}

  /**
   * Class representing the Users collection in Thingworx.
   */
  class Users extends EntityCollection {
    /**
     * Creates the Users collection of a particular Thingworx server.
     *
     * @param {Server} server - the Thingworx server to which this collection belongs.
     */
    constructor(server) {
      super('Users', User, server);
    }
  }

  /**
   * Class representing a Thingworx server.
   * Use the .getService() method instead of the constructor for creating instances of this class.
   */
  class Server {
    /**
     * Creates a reference to a Thingworx server.
     * Precondition: the server origin should already have been cleaned using the static method
     * cleanServerOrigin of this class.
     * Do not call this directly. Use the static method .getService() of this class instead.
     *
     * @param {string} origin - the server origin.
     */
    constructor(origin) {
      this.origin = origin;

      // collections - these fields will be lazily populated when a collection gets accessed using
      // the getters defined below. opt stands for optional, which means that the property will be
      // null until it is accessed for the first time.
      this.opt_logs = null;
      this.opt_users = null;
      this.opt_dashboards = null;
      this.opt_groups = null;
      this.opt_mashups = null;
      this.opt_menus = null;
      this.opt_modelTags = null;
      this.opt_networks = null;
      this.opt_dataTags = null;
      this.opt_dataShapes = null;
      this.opt_thingShapes = null;
      this.opt_thingTemplates = null;
      this.opt_things = null;
      this.opt_resources = null;
    }

    /**
     * Gets the Logs collection
     *
     * @returns {EntityCollection} - the Logs collection
     */
    get logs() {
      if (this.opt_logs === null) {
        this.opt_logs = new Logs(this);
      }
      return this.opt_logs;
    }

    /**
     * Gets the Users collection
     *
     * @returns {EntityCollection} - the Users collection
     */
    get users() {
      if (this.opt_users === null) {
        this.opt_users = new Users(this);
      }
      return this.opt_users;
    }

    /**
     * Gets the Dashboards collection
     *
     * @returns {EntityCollection} - the Dashboards collection
     */
    get dashboards() {
      if (this.opt_dashboards === null) {
        this.opt_dashboards = new Dashboards(this);
      }
      return this.opt_dashboards;
    }

    /**
     * Gets the Groups collection
     *
     * @returns {EntityCollection} - the Groups collection
     */
    get groups() {
      if (this.opt_groups === null) {
        this.opt_groups = new Groups(this);
      }
      return this.opt_groups;
    }

    /**
     * Gets the Menus collection
     *
     * @returns {EntityCollection} - the Menus collection
     */
    get menus() {
      if (this.opt_menus === null) {
        this.opt_menus = new Menus(this);
      }
      return this.opt_menus;
    }

    /**
     * Gets the Mashups collection
     *
     * @returns {EntityCollection} - the Mashups collection
     */
    get mashups() {
      if (this.opt_mashups === null) {
        this.opt_mashups = new Mashups(this);
      }
      return this.opt_mashups;
    }

    /**
     * Gets the ModelTags collection
     *
     * @returns {EntityCollection} - the ModelTags collection
     */
    get modelTags() {
      if (this.opt_modelTags === null) {
        this.opt_modelTags = new ModelTags(this);
      }
      return this.opt_modelTags;
    }

    /**
     * Gets the Networks collection
     *
     * @returns {EntityCollection} - the Networks collection
     */
    get networks() {
      if (this.opt_networks === null) {
        this.opt_networks = new Networks(this);
      }
      return this.opt_networks;
    }

    /**
     * Gets the DataTags collection
     *
     * @returns {EntityCollection} - the DataTags collection
     */
    get dataTags() {
      if (this.opt_dataTags === null) {
        this.opt_dataTags = new DataTags(this);
      }
      return this.opt_dataTags;
    }

    /**
     * Gets the DataShapes collection
     *
     * @returns {EntityCollection} - the DataShapes collection
     */
    get dataShapes() {
      if (this.opt_dataShapes === null) {
        this.opt_dataShapes = new DataShapes(this);
      }
      return this.opt_dataShapes;
    }

    /**
     * Gets the ThingShapes collection
     *
     * @returns {EntityCollection} - the ThingShapes collection
     */
    get thingShapes() {
      if (this.opt_thingShapes === null) {
        this.opt_thingShapes = new ThingShapes(this);
      }
      return this.opt_thingShapes;
    }

    /**
     * Gets the ThingTemplates collection
     *
     * @returns {EntityCollection} - the ThingTemplates collection
     */
    get thingTemplates() {
      if (this.opt_thingTemplates === null) {
        this.opt_thingTemplates = new ThingTemplates(this);
      }
      return this.opt_thingTemplates;
    }

    /**
     * Gets the Things collection
     *
     * @returns {EntityCollection} - the Things collection
     */
    get things() {
      if (this.opt_things === null) {
        this.opt_things = new Things(this);
      }
      return this.opt_things;
    }

    /**
     * Gets the Resources collection
     *
     * @returns {EntityCollection} - the Resources collection
     */
    get resources() {
      if (this.opt_resources === null) {
        this.opt_resources = new Resources(this);
      }
      return this.opt_resources;
    }

    /**
     * Creates a reference to a Thingworx server or, if a reference to that server was already
     * created, returns that one instead of creating a new one.
     * Prefer calling this method in place of the constructor.
     *
     * @param {string} serverOrigin - the server origin.
     * @returns {Server} - an instance of Server associated to the given server origin.
     */
    static getServer(serverOrigin) {
      const { cache } = Server;
      const cleanedServerOrigin = Server.cleanServerOrigin(serverOrigin);
      if (!cache.has(cleanedServerOrigin)) {
        cache.set(cleanedServerOrigin, new Server(cleanedServerOrigin));
      }
      return cache.get(cleanedServerOrigin);
    }

    /**
     * Cleans server origin by adding the http scheme (if no scheme is present) and removing the
     * trailing slash (if present).
     * The empty string is not cleaned, it is returned as it is.
     *
     * @param {string} serverOrigin - the server origin.
     * @returns {string} the cleaned server origin.
     */
    static cleanServerOrigin(serverOrigin) {
      if (serverOrigin === '') return serverOrigin;

      const serverOriginWithScheme = serverOrigin.includes('://') ? serverOrigin : `http://${serverOrigin}`;
      return serverOriginWithScheme.replace(/\/$/, '');
    }
  }

  /**
   * Used for caching server instances. This avoids creating new instances that point to the same
   * server.
   */
  Server.cache = new Map();

  /**
   * Class that handles the creation of proxies intended for easy usage.
   * Proxies allow to call from the browser Javascript Thingworx services in a Thingworx-ish way.
   * For example,
   *
   * const result = Things['MyThing'].MyService()
   *
   * even though there is no 'MyThing' property in the Things collection, nor a .MyService() method
   * in the 'MyThing' object.
   *
   * Proxies are associated to a particular connection instance. Explanation:
   * A server contains entity collections, which contains entities, which contain services, which can
   * be called. The link is like this:
   *
   *          |---> Things------------>|---> Thing 1 --->|---> Service 1 --> (services are called by
   *          |---> Things Shapes      |---> Thing 2     |---> Service 2     passing them a connection
   * Server-->|---> Things Templates   |---> Thing 3     |---> Service 3     instance and an object
   *          |---> .......            |---> .......     |---> .......       with its parameters)
   *          |---> .......            |---> .......     |---> .......
   *
   * This model is independent of proxies whatsoever.
   * A connection is a defined as a pair of a server and authentication parameters for that server.
   * A server instance can exist, and there can be many connection instances that connect,
   * authenticate to, and make requests to that server.
   *
   * How does the user can use the Connection class to call entity services of a Thingworx server?
   *
   * This is where proxies como into play.
   *
   * Proxies create entry points for calling the Thingworx services from a particular connection. So
   * once you create a connection, you can create proxies associated to that connection. Then, every
   * service call through the proxies will be sent using the connection object to which the proxies
   * belong.
   *
   * You can create proxies for a connection using the .createEntityCollectionProxies(connection)
   * method.
   */
  class ProxyCreator {
    /**
     * Creates proxies of all entity collections for a connection instance.
     *
     * @param {Connection} conn - The connection instance for which to build the proxies.
     * @returns {object} a frozen object containing all proxies.
     */
    static createEntityCollectionProxies(conn) {
      const { server } = conn;

      let logs = null;
      let users = null;
      let dashboards = null;
      let groups = null;
      let mashups = null;
      let menus = null;
      let modelTags = null;
      let networks = null;
      let dataTags = null;
      let dataShapes = null;
      let thingShapes = null;
      let thingTemplates = null;
      let things = null;
      let resources = null;

      return Object.freeze({
        get Logs() {
          if (logs === null) {
            logs = ProxyCreator.createCollectionProxy(server.logs, conn);
          }
          return logs;
        },
        get Users() {
          if (users === null) {
            users = ProxyCreator.createCollectionProxy(server.users, conn);
          }
          return users;
        },
        get Dashboards() {
          if (dashboards === null) {
            dashboards = ProxyCreator.createCollectionProxy(server.dashboards, conn);
          }
          return dashboards;
        },
        get Groups() {
          if (groups === null) {
            groups = ProxyCreator.createCollectionProxy(server.groups, conn);
          }
          return groups;
        },
        get Mashups() {
          if (mashups === null) {
            mashups = ProxyCreator.createCollectionProxy(server.mashups, conn);
          }
          return mashups;
        },
        get Menus() {
          if (menus === null) {
            menus = ProxyCreator.createCollectionProxy(server.menus, conn);
          }
          return menus;
        },
        get ModelTags() {
          if (modelTags === null) {
            modelTags = ProxyCreator.createCollectionProxy(server.modelTags, conn);
          }
          return modelTags;
        },
        get Networks() {
          if (networks === null) {
            networks = ProxyCreator.createCollectionProxy(server.networks, conn);
          }
          return networks;
        },
        get DataTags() {
          if (dataTags === null) {
            dataTags = ProxyCreator.createCollectionProxy(server.dataTags, conn);
          }
          return dataTags;
        },
        get DataShapes() {
          if (dataShapes === null) {
            dataShapes = ProxyCreator.createCollectionProxy(server.dataShapes, conn);
          }
          return dataShapes;
        },
        get ThingShapes() {
          if (thingShapes === null) {
            thingShapes = ProxyCreator.createCollectionProxy(server.thingShapes, conn);
          }
          return thingShapes;
        },
        get ThingTemplates() {
          if (thingTemplates === null) {
            thingTemplates = ProxyCreator.createCollectionProxy(server.thingTemplates, conn);
          }
          return thingTemplates;
        },
        get Things() {
          if (things === null) {
            things = ProxyCreator.createCollectionProxy(server.things, conn);
          }
          return things;
        },
        get Resources() {
          if (resources === null) {
            resources = ProxyCreator.createCollectionProxy(server.resources, conn);
          }
          return resources;
        },
        // For changing the authentication params.
        changeAuthMethod(authParams) {
          conn.setAuthParams(authParams);
        },
      });
    }

    /**
     * Creates a proxy for a particular entity collection. As new entities of this particular entity
     * collection get used, they will be remembered by the proxy, so the next time an entity gets
     * used, the proxy will use its existing reference instead of creating a new one.
     *
     * @param {EntityCollection} entityCollection - the collection for which to create a proxy.
     * @param {Connection} connection - The connection instance to which this proxy belongs to. In
     * other words, the connection instance to use to making requests from the entities.
     * @returns {Proxy} the proxy for the entity collection.
     */
    static createCollectionProxy(entityCollection, connection) {
      const proxyHandler = {
        get(entityProxies, entityName) {
          if (!entityProxies.has(entityName)) {
            const entity = entityCollection.getEntity(entityName);
            const entityProxy = ProxyCreator.createEntityProxy(entity, connection);
            entityProxies.set(entityName, entityProxy);
          }
          return entityProxies.get(entityName);
        },
      };

      const entityProxies = new Map();
      return Object.freeze(new Proxy(entityProxies, proxyHandler));
    }

    /**
     * Creates a proxy for a particular entity. As new services of this particular entity get called,
     * they will be remembered by the proxy, so the next time a service is called, the proxy will use
     * its existing reference instead of creating a new one.
     *
     * @param {Entity} entity - the entity for which to create a proxy.
     * @param {Connection} conn - The connection instance to which this proxy belongs to. In other
     * words, the connection instance to use when calling the entities services.
     * @returns {Proxy} the entity proxy.
     */
    static createEntityProxy(entity, conn) {
      const proxyHandler = {
        get(propertyOrServiceProxies, propOrServiceName) {
          if (!propertyOrServiceProxies.has(propOrServiceName)) {
            const proxy = ProxyCreator.createPropertyOrServiceProxy(propOrServiceName, entity, conn);
            propertyOrServiceProxies.set(propOrServiceName, proxy);
          }
          return propertyOrServiceProxies.get(propOrServiceName);
        },
      };

      const propertyOrServiceProxies = new Map();
      return Object.freeze(new Proxy(propertyOrServiceProxies, proxyHandler));
    }

    /**
     * Creates a proxy for an entity's property or service.
     *
     * @param {string} propertyOrServiceName - the property or service to proxy.
     * @param {Entity} entity - the entity which owns the property or service to proxy.
     * @param {Connection} connection - the connection to use for the requests associated to the
     * property or service.
     * @returns {Proxy} - the proxy for the property or service.
     */
    static createPropertyOrServiceProxy(propertyOrServiceName, entity, connection) {
      const proxyHandler = {
        // This applies if propertyOrService is a property and executes when calling .then(),
        // .val(), .json(), .infoTable(), or .set() methods. Any other method results in an error.
        get(_, methodName) {
          switch (methodName) {
            case 'then':
            case 'catch':
            case 'finally':
              return (...args) => entity.getProperty(propertyOrServiceName)
                .get(connection)[methodName](...args);
            case 'set':
              return async (value) => {
                const response = await entity.getProperty(propertyOrServiceName)
                  .set(value, connection);
                if (!response.ok) {
                  throw await response.buildError();
                }
              };
            default:
              throw new Error(`Unrecognized method ${methodName}`);
          }
        },
        // This will interpret that propertyOrService is a service and execute when appending ()
        // next to the service name.
        apply(_, thisArg, [params]) {
          return entity.getService(propertyOrServiceName).call(params, connection);
        },
      };
      // using a function as the proxy target to enable the 'apply' trap.
      return Object.freeze(new Proxy(() => {}, proxyHandler));
    }
  }

  /**
   * Checks if an object has no properties.
   *
   * @param {object} obj - the object to check.
   * @returns {boolean} true if the object is empty, false otherwise.
   */
  function isObjectEmpty(obj) {
    return Object.entries(obj).length === 0 && obj.constructor === Object;
  }

  /**
   * The request options modifier will modify the request options before sending out the request
   * to the Thingworx server. This will ensure that the options contains the correct authentication
   * parameters.
   */
  class OptionsModifier {
    /**
     * Sets the credentials to 'include' in the request options.
     *
     * @param {object} options - the options that will be used for a fetch request.
     */
    static addCredentialsToOptions(options) {
      options.credentials = 'include';
    }

    /**
     * Returns the options modifier callback for when the authentication mode is 'appKey'.
     *
     * @param {string} appKey - the appKey to use for authentication.
     * @returns {Function} the options modifier callback for when the authentication mode is 'appKey'.
     */
    static createAppKeyModifier(appKey) {
      return function addAppKeyToOptions(options) {
        if (options.headers == null) {
          options.headers = {};
        }
        options.headers.appKey = appKey;
      };
    }

    /**
     * Returns the options modifier callback for when the authentication mode is 'basic'.
     *
     * @param {string} username - the username to use for authentication.
     * @param {string} password - the password to use for authentication.
     * @returns {Function} the options modifier callback for when the authentication mode is 'basic'.
     */
    static createBasicAuthModifier(username, password) {
      const authHeader = `Basic ${btoa(`${username}:${password}`)}`;
      return function addBasicAuthToOptions(options) {
        if (options.headers == null) {
          options.headers = {};
        }
        options.headers.Authorization = authHeader;
      };
    }

    /**
     * Returns the request options modifier customized according to the authentication parameters.
     * Caches the existing modifiers (the ones that were returned before through this method), so
     * existing ones are returned instead of creating and duplicating new modifiers.
     *
     * @param {object} authParams - the authentication parameters used to authenticate to the
     * Thingworx server. See the documentation of Thingworx.collections() or Thingworx.connect() for
     * details of these parameters.
     * @returns {Function} the modifier function that will modify the options of the requests to
     * include the authentication parameters given
     */
    static getModifier(authParams) {
      if (authParams == null || isObjectEmpty(authParams)) {
        return OptionsModifier.addCredentialsToOptions;
      }

      const { appKey, username, password } = authParams;

      // Check if we should return an app key modifier
      const { appKeyModifiersCache } = OptionsModifier;
      if (appKey != null && username == null && password == null) {
        if (!appKeyModifiersCache.has(appKey)) {
          const modifier = OptionsModifier.createAppKeyModifier(appKey);
          appKeyModifiersCache.set(appKey, modifier);
        }
        return appKeyModifiersCache.get(appKey);
      }

      // Check if we should return a basic authorization modifier
      const { basicAuthModifiersCache } = OptionsModifier;
      if (username != null && password != null && appKey == null) {
        const key = `${btoa(`${username}:${password}`)}`;
        if (!basicAuthModifiersCache.has(key)) {
          const modifier = OptionsModifier.createBasicAuthModifier(username, password);
          basicAuthModifiersCache.set(key, modifier);
        }
        return basicAuthModifiersCache.get(key);
      }

      // throws an error because auth parameters combination was wrong.
      throw new Error('Authentication parameters combination is wrong.');
    }
  }

  /** Maps appKeys to modifiers */
  OptionsModifier.appKeyModifiersCache = new Map();

  /** Maps username/password combinations to modifiers */
  OptionsModifier.basicAuthModifiersCache = new Map();

  /**
   * Class representing a Login response.
   */
  class PingResponse extends ThingworxResponse {}

  /**
   * Class representing a ping to the Thingworx server.
   */
  class Ping extends RequestNeedingAuthentication {
    /**
     * Creates a ping to the Thingworx server.
     *
     * @param {string} server - the Thingworx server to ping. Instance of Server.
     */
    constructor(server) {
      const url = Ping.buildRequestURL(server);
      const options = null;
      super(url, options, server);
    }

    /**
     * Builds the request URL for the pings
     *
     * @param {Server} server - the Server instance for which to build the URL.
     * @returns {string} the request URL.
     */
    static buildRequestURL(server) {
      return `${server.origin}/Thingworx/Composer`;
    }

    /**
     * Sends a ping to the Thingworx server.
     *
     * @param {Connection} connection - the connection instance to use to make the request.
     * @returns {Promise} a promise that resolves to an instance of PingResponse.
     */
    async send(connection) {
      const response = await connection.sendRequest(this);
      return new PingResponse(response);
    }
  }

  /**
   * Class used to create a connection to a particular Thingworx server.
   * A connection is defined by a Server and authentication parameters to that server.
   * Once a connection is created, you can use its collection proxies (located on the
   * collectionProxies property) to call Thingworx services.
   */
  class Connection {
    /**
     * Creates a new connection to a particular Thingworx server.
     *
     * @param {string|Server} server - the Thingworx server for which to create a connection.
     * @param {object} authParams - the authentication parameters used to authenticate to the
     * Thingworx server. See the documentation of Thingworx.collections() or Thingworx.connect() for
     * details of these parameters.
     */
    constructor(server, authParams) {
      if (typeof server === 'string') {
        this.server = Server.getServer(server);
      } else if (server.constructor.name === 'Server') {
        this.server = server;
      } else {
        throw new Error('Server parameter should be either a string or an instance of the Server class.');
      }
      this.modifyOptions = OptionsModifier.getModifier(authParams);
      this.collectionProxies = ProxyCreator.createEntityCollectionProxies(this);
    }

    /**
     * Sends a request to the Thingworx server, modifying it to make sure it contains the correct
     * authorization parameters.
     *
     * @param {RequestNeedingAuthentication} request - the request to send to the Thingworx server.
     * This request is missing authentication parameters.
     * @returns {Promise} - a promise resulting from the fetch operation.
     */
    async sendRequest(request) {
      if (request.options == null) {
        request.options = {};
      }
      this.modifyOptions(request.options);
      return fetch(request.url, request.options);
    }

    /**
     * Pings the server and asynchronously waits for a response.
     *
     * @returns {Promise} a promise that resolves to an instance of PingResponse.
     */
    async ping() {
      const request = new Ping(this.server);
      return request.send(this);
    }

    /**
     * Sets new authentication parameters with which to make requests.
     *
     * @param {object} authParams - the new authentication parameters used to communicate to the
     * Thingworx server. See the documentation of Thingworx.collections() or Thingworx.connect() for
     * details of these parameters.
     */
    setAuthParams(authParams) {
      this.modifyOptions = OptionsModifier.getModifier(authParams);
    }
  }

  /**
   * Entry class of the library.
   */
  class Thingworx {
    /**
     * Create references to entity collections in a particular Thingworx server.
     * Entity collections returned are: Dashboards, DataShapes, DataTags, Groups, Logs, Mashups,
     * Menus, ModelTags, Networks, Resources, Things, ThingShapes, ThingTemplates, and Users.
     *
     * Unlike the .connect() method, this method does not check the validity of the authentication
     * parameters before making the requests. Prefer this method over .connect().
     *
     * @param {string} [serverOrigin] - the Thingworx server to connect to. Defaults to the empty
     * string, which will be considered as the current origin (window.origin). If no scheme (protocol)
     * is in the origin url, then an 'http' protocol will be prepended by default.
     * @param {object} [authParams] - the parameters to use to authenticate to the Thingworx server
     * when doing the requests. If not given, the browser will try to use credentials (cookies,
     * authorization headers, or TLS certificates) for the HTTP requests. This is the same as setting
     * the credentials option to 'include' in the Requests.
     * @param {string} [authParams.appKey] - the appKey to use while making the REST API calls.
     * @param {string} [authParams.username] - the username to use to authenticate with Basic
     * authorization.
     * @param {string} [authParams.password] - the password to use to authenticate with Basic
     * authorization.
     *
     * Only one of the appKey or the combination of the username and password can be set. The other
     * one must forcefully be undefined, null, or not present in the authentication parameters object.
     * Prefer to use appKey or the browser credentials over username/password. Requests are much
     * faster using these methods.
     *
     * @returns {object} references to all entity collections wrapped in an object. Use object
     * destructuring to access those references. See examples.
     *
     * @example
     * const { Things } = Thingworx.collections('http://localhost:8080');
     * @example
     * const { ThingShapes, Mashups } = Thingworx.collections('localhost', {
     *   appKey: '896548f2-eaab-46a4-b129-53f1531557a4'
     * });
     * @example
     * const { Resources, Networks } = Thingworx.collections('https://localhost:8080', {
     *   username: 'admin',
     *   password: 'myPassw0rd'
     * });
     */
    static collections(serverOrigin = '', authParams) {
      const connection = new Connection(serverOrigin, authParams);
      return connection.collectionProxies;
    }

    /**
     * Authenticate to a Thingworx server and, if successful, create references to entity collections
     * from it.
     * Entity collections returned are: Dashboards, DataShapes, DataTags, Groups, Logs, Mashups,
     * Menus, ModelTags, Networks, Resources, Things, ThingShapes, ThingTemplates, and Users.
     *
     * Unlike the .collections() method, this method checks the validity of the authentication
     * parameters before returning the entity collection references. Prefer the .collections() method
     * over this method.
     *
     * @param {string} [serverOrigin] - the Thingworx server to connect to. Defaults to the empty
     * string, which will be considered as the current origin (window.origin). If no scheme (protocol)
     * is in the origin url, then an 'http' protocol will be prepended by default.
     * @param {object} [authParams] - the parameters to use to authenticate to the Thingworx server
     * when doing the requests. If not given, the browser will try to use credentials (cookies,
     * authorization headers, or TLS certificates) for the HTTP requests. This is the same as setting
     * the credentials option to 'include' in the Requests.
     * @param {string} [authParams.appKey] - the appKey to use while making the REST API calls.
     * @param {string} [authParams.username] - the username to use to authenticate with Basic
     * authorization.
     * @param {string} [authParams.password] - the password to use to authenticate with Basic
     * authorization.
     *
     * Only one of the appKey or the combination of the username and password can be set. The other
     * one must forcefully be undefined, null, or not present in the authentication parameters object.
     * Prefer to use appKey or the browser credentials over username/password. Requests are much
     * faster using these methods.
     *
     * @returns {object} references to all entity collections wrapped in an object. Use object
     * destructuring to access those references. See examples.
     * @throws {Error} an error if the authentication parameters are invalid or the request couldn't
     * be completed successfully
     *
     * In order for the examples to work, they should be inside an async function.
     *
     * @example
     * const { Things } = await Thingworx.connect('http://localhost:8080')
     * @example
     * const { ThingShapes, Mashups } = await Thingworx.connect('localhost', {
     *   appKey: '896548f2-eaab-46a4-b129-53f1531557a4'
     * });
     * @example
     * const { Resources, Networks } = await Thingworx.connect('https://localhost:8080', {
     *   username: 'admin',
     *   password: 'myPassw0rd'
     * });
     */
    static async connect(serverOrigin = '', authParams) {
      const connection = new Connection(serverOrigin, authParams);
      const pingResponse = await connection.ping();
      if (!pingResponse.ok) {
        throw pingResponse.buildDefaultConnectionError();
      }
      return connection.collectionProxies;
    }
  }

  return Thingworx;

})));
